/************************************************************************
 *   IRC - Internet Relay Chat, src/m_gline.c
 *   Copyright (C) 1990 Jarkko Oikarinen and
 *                      University of Oulu, Computing Center
 *
 *   See file AUTHORS in IRC package for additional names of
 *   the programmers. 
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  $Id: m_gline.c,v 1.2 2001/02/10 19:00:43 jpinto Exp $
 */
#include "m_gline.h"
#include "channel.h"
#include "client.h"
#include "common.h"
#include "config.h"
#include "dline_conf.h"
#include "irc_string.h"
#include "ircd.h"
#include "m_kline.h"
#include "mtrie_conf.h"
#include "numeric.h"
#include "s_conf.h"
#include "s_misc.h"
#include "s_serv.h"
#include "scache.h"
#include "send.h"
#include "struct.h"

#include <assert.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <fcntl.h>



/* external variables */
extern ConfigFileEntryType ConfigFileEntry; /* defined in ircd.c */

/* internal variables */
aConfItem *glines = (aConfItem *)NULL;
extern aConfItem *glines;

/* internal functions */
static void add_gline(aConfItem *);
#if 0
static void log_gline_request(const char*,const char*,const char*,
                              const char* oper_server,
                              const char *,const char *,const char *);


static int majority_gline(aClient*, const char *,const char *, const char *, 
                          const char* serv_name,
                          const char *,const char *,const char *); 
#endif /* 0 */

/*
 * m_gline()
 *
 *		parv[0] = sender prefix
 *		parv[1] = user@host
 *		parv[2] = time
 *		parv[3] = reason (if !MyClient its Sender)
 *		parv[4] = reason (if !MyClient)  
 */
int     m_gline(aClient *cptr,
                aClient *sptr,
                int parc,
                char *parv[])
{
  char *oper_name;              /* nick of oper requesting GLINE */
  char *oper_username;          /* username of oper requesting GLINE */
  char *oper_host;              /* hostname of oper requesting GLINE */
  const char* oper_server;      /* server of oper requesting GLINE */
  char *user, *host;            /* user and host of GLINE "victim" */
  char *reason;                 /* reason for "victims" demise */
  char *who;					/* who added the gline */
  char *p;
  register char tmpch;
  register int nonwild;
  char buffer[512];
  const char *current_date;
  char tempuser[USERLEN + 2];
  char temphost[HOSTLEN + 1];
  char *ch, *timestr;
  unsigned long secs;
  aConfItem *aconf;
  
  if(MyClient(sptr)) /* allow remote opers to apply g lines */
    {

      /* Only globals can apply Glines */
      if (!IsOper(sptr))
        {
          sendto_one(sptr, form_str(ERR_NOPRIVILEGES), me.name, parv[0]);
          return 0;
        }

      if (!IsSetOperGline(sptr))
        {
          sendto_one(sptr,":%s NOTICE %s :You have no G flag",me.name,parv[0]);
          return 0;
        }

      if ( parc < 4 )
        {
          sendto_one(sptr, form_str(ERR_NEEDMOREPARAMS),
                     me.name, parv[0], "GLINE");
          return 0;
        }
	}

  if ( (host = strchr(parv[1], '@')) || *parv[1] == '*' )
    {
	    /* Explicit user@host mask given */
		
        if(host)                      /* Found user@host */
      	  {
        	user = parv[1];   /* here is user part */
        	*(host++) = '\0'; /* and now here is host */
      	  }
    	else
      	  {
        	user = "*";               /* no @ found, assume its *@somehost */
        	host = parv[1];
      	  }

    	if (!*host)           /* duh. no host found, assume its '*' host */
      	  host = "*";
  
    	strncpy_irc(tempuser, user, USERLEN + 1);     /* allow for '*' */
  		tempuser[USERLEN + 1] = '\0';
	    strncpy_irc(temphost, host, HOSTLEN);
  		temphost[HOSTLEN] = '\0';
	    user = tempuser;
  		host = temphost;
  	  }
  else
    {
	  sendto_one(sptr, ":%s NOTICE %s :Can't G-Line a nick use user@host",
           me.name,
           parv[0]);
    	return 0;
  	}
					
	 /*
	  * Now we must check the user and host to make sure there
  	  * are at least NONWILDCHARS non-wildcard characters in
  	  * them, otherwise assume they are attempting to gline
      * *@* or some variant of that. This code will also catch
      * people attempting to gline *@*.tld, as long as NONWILDCHARS
      * is greater than 3. In that case, there are only 3 non-wild
      * characters (tld), so if NONWILDCHARS is 4, the gline will
      * be disallowed.
	  * -wnder
  	  */

	nonwild = 0;
	p = user;
	while ((tmpch = *p++))
	  {
		if (!IsKWildChar(tmpch))
  	  	  {
	      /*
  	       * If we find enough non-wild characters, we can
    	   * break - no point in searching further.
	       */
  			if (++nonwild >= NONWILDCHARS)
      		break;
  		  }
	  }

	if (nonwild < NONWILDCHARS)
	  {
  	   /*
    	* The user portion did not contain enough non-wild
  	    * characters, try the host.
    	*/
  		p = host;
  		while ((tmpch = *p++))
  		  {
    	    if (!IsKWildChar(tmpch))
      		  if (++nonwild >= NONWILDCHARS)
        		break;
  		  }
	  }
	  
	if (nonwild < NONWILDCHARS)
	  {
	  	  /*
		   * Not enough non-wild characters were found, assume
    	   * they are trying to gline *@*.
    	   */
  		if (MyClient(sptr))
    	  sendto_one(sptr,
      		":%s NOTICE %s :Please include at least %d non-wildcard characters with the user@host",
      		me.name,
      		parv[0],
      		NONWILDCHARS);
  		return 0;
	  }		
	  who = IsServer(cptr) ? parv[3]: sptr->name;
	  reason = IsServer(cptr) ? parv[4] : parv[3];

#ifdef OLD_TS  
  if(IsServer(sptr) && !IsCapable(sptr,CAP_PTS4))
	{
	  who = parv[4];
	  reason = parv[5];
	} 
#endif

  if(strchr(reason, ':'))
	{
  	  sendto_one(sptr,
    	":%s NOTICE %s :Invalid character ':' in comment",
         me.name, parv[2]);
        return 0;
    }

  timestr= parv[2];  
  ch=&timestr[strlen(timestr)-1];
  switch(*ch) 
	{
  	  case 'm': secs = atol(timestr)*60;break;
      case 'h': secs = atol(timestr)*3600;break;
      case 'd': secs = atol(timestr)*3600*24;break;
      default : secs = atol(timestr);break;
    } 	

  if (find_is_glined(host, user)) {
	if(MyClient(sptr))
	    sendto_one(sptr,
		   ":%s NOTICE %s :*** Cannot add G:Line - Match already exists!", 
		   me.name, sptr->name);
	return 0;
  }
  
  aconf = make_conf();
  aconf->status = CONF_KILL;

  
  if( MyClient(sptr) )
	{
	  current_date = smalldate((time_t) 0);
	  ircsprintf(buffer, "%s (%s)",reason,current_date);
	} else
	  ircsprintf(buffer, "%s", reason);
	  
  DupString(aconf->host, host);      
  DupString(aconf->passwd, buffer);
  DupString(aconf->user, user);
  DupString(aconf->name, who);
  aconf->hold = CurrentTime + secs;
  
  add_gline(aconf);

  if(IsClient(sptr)) 
	{
  
  	  if (sptr->user && sptr->user->server)
  		oper_server = sptr->user->server;
	  else
  		return 0;
	  
	  oper_name     = sptr->name;
	  oper_username = sptr->username;
	  oper_host     = sptr->host;
	  
	  sendto_ops("%s!%s@%s on %s added gline for \2[\2%s@%s\2]\2 \2[\2%s\2]\2 expiring on %s",
		oper_name,
  		oper_username,
  		oper_host,
  		oper_server,
  		user,
  		host,
  		reason,
		smalldate(aconf->hold));  	
	  
	}
  sendto_match_cap_servs(NULL, IsServer(cptr) ? cptr : NULL, CAP_PTS4,
  						":%s GLINE %s@%s %lu %s :%s",
                         sptr->name,
                         user,
                         host,
						 secs,
						 who,
                         reason);

#ifdef OLD_TS
  sendto_nomatch_cap_servs(NULL, IsServer(cptr) ? cptr : NULL, CAP_PTS4,
  						":%s GLINE %s@%s %lu %lu %s :%s",
                         me.name,
                         user,
                         host,
						 secs,
						 CurrentTime,	 
						 who,
                         reason);
#endif						 						 
  rehashed = YES;
  dline_in_progress = NO;
  nextping = CurrentTime;

  return 0;
}

#if 0
/*
 * log_gline_request()
 *
 */
static void log_gline_request(
                      const char *oper_nick,
                      const char *oper_user,
                      const char *oper_host,
                      const char* oper_server,
                      const char *user,
                      const char *host,
                      const char *reason)
{
  char        buffer[1024];
  char        filenamebuf[PATH_MAX + 1];
  static char timebuffer[MAX_DATE_STRING];
  struct tm*  tmptr;
  int         out;

  ircsprintf(filenamebuf, "%s.%s", 
             ConfigFileEntry.glinefile, small_file_date((time_t)0));
  if ((out = open(filenamebuf, O_RDWR|O_APPEND|O_CREAT,0644))==-1)
    {
      sendto_realops("*** Problem opening %s",filenamebuf);
      return;
    }

  tmptr = localtime(&CurrentTime);
  strftime(timebuffer, MAX_DATE_STRING, "%Y/%m/%d %H:%M:%S", tmptr);

  ircsprintf(buffer,
           "#Gline for %s@%s [%s] requested by %s!%s@%s on %s at %s\n",
           user,host,reason,
           oper_nick,oper_user,oper_host,oper_server,
           timebuffer);

  if (write(out, buffer, strlen(buffer)) <= 0)
    {
      sendto_realops("*** Problem writing to %s",filenamebuf);
    }
  close(out);
}


/*
 * log_gline()
 *
 */
static void log_gline(aClient *sptr,
                      const char *parv0,
                      GLINE_PENDING *gline_pending_ptr,
                      const char *oper_nick,
                      const char *oper_user,
                      const char *oper_host,
                      const char *oper_server,
                      const char *user,
                      const char *host,
                      const char *reason)
{
  char         buffer[1024];
  char         filenamebuf[PATH_MAX + 1];
  static  char timebuffer[MAX_DATE_STRING + 1];
  struct tm*   tmptr;
  int          out;

  ircsprintf(filenamebuf, "%s.%s", 
                ConfigFileEntry.glinefile, small_file_date((time_t) 0));
  if ((out = open(filenamebuf, O_RDWR|O_APPEND|O_CREAT,0644))==-1)
    {
      return;
    }

  tmptr = localtime(&CurrentTime);
  strftime(timebuffer, MAX_DATE_STRING, "%Y/%m/%d %H:%M:%S", tmptr);

  ircsprintf(buffer,"#Gline for %s@%s %s added by the following\n",
                   user,host,timebuffer);

  /*
   * NOTE: safe_write closes the file if an error occurs
   */
  if (safe_write(sptr, filenamebuf, out, buffer))
    return;

  ircsprintf(buffer, "#%s!%s@%s on %s [%s]\n",
                   gline_pending_ptr->oper_nick1,
                   gline_pending_ptr->oper_user1,
                   gline_pending_ptr->oper_host1,
                   gline_pending_ptr->oper_server1,
                   (gline_pending_ptr->reason1)?
                   (gline_pending_ptr->reason1):"No reason");

  if (safe_write(sptr,filenamebuf,out,buffer))
    return;

  ircsprintf(buffer, "#%s!%s@%s on %s [%s]\n",
                   gline_pending_ptr->oper_nick2,
                   gline_pending_ptr->oper_user2,
                   gline_pending_ptr->oper_host2,
                   gline_pending_ptr->oper_server2,
                   (gline_pending_ptr->reason2)?
                   (gline_pending_ptr->reason2):"No reason");

  if (safe_write(sptr, filenamebuf, out, buffer))
    return;

  ircsprintf(buffer, "#%s!%s@%s on %s [%s]\n",
                   oper_nick,
                   oper_user,
                   oper_host,
                   oper_server,
                   (reason)?reason:"No reason");

  if (safe_write(sptr,filenamebuf,out,buffer))
    return;

  ircsprintf(buffer, "K:%s:%s:%s\n", host,user,reason);
  if (safe_write(sptr,filenamebuf,out,buffer))
    return;

  close(out);
}
#endif

/*
 * flush_glines
 * 
 * inputs       - NONE
 * output       - NONE
 * side effects -
 *
 * Get rid of all placed G lines, hopefully to be replaced by gline.log
 * placed k-lines
 */
void flush_glines()
{
  aConfItem *kill_list_ptr;

  if((kill_list_ptr = glines))
    {
      while(kill_list_ptr)
        {
          glines = kill_list_ptr->next;
          free_conf(kill_list_ptr);
          kill_list_ptr = glines;
        }
    }
  glines = (aConfItem *)NULL;
}

/* find_gkill
 *
 * inputs       - aClient pointer to a Client struct
 * output       - aConfItem pointer if a gline was found for this client
 * side effects - none
 */

aConfItem *find_gkill(aClient* cptr, char* username)
{
  aConfItem *foundgline;
  assert(0 != cptr);
  
  if (IsElined(cptr)||IsExemptGline(cptr))
	return (aConfItem *) NULL;
	
  if (!(foundgline = find_is_glined(cptr->realhost, username)) && 
	  !(foundgline = find_is_glined(cptr->host, username)))
    foundgline = (aConfItem *) NULL;
	
  return (foundgline);
}

/*
 * find_is_glined
 * inputs       - hostname
 *              - username
 * output       - pointer to aConfItem if user@host glined
 * side effects -
 *  WARNING, no sanity checking on length of name,host etc.
 * thats expected to be done by caller.... *sigh* -Dianora
 */

struct ConfItem* find_is_glined(const char* host, const char* username)
{
  aConfItem *kill_list_ptr;     /* used for the link list only */
  aConfItem *last_list_ptr;
  aConfItem *tmp_list_ptr;

  /* gline handling... exactly like temporary klines 
   * I expect this list to be very tiny. (crosses fingers) so CPU
   * time in this, should be minimum.
   * -Dianora
  */

  if(glines)
    {
      kill_list_ptr = last_list_ptr = glines;

      while(kill_list_ptr)
        {
          if(kill_list_ptr->hold <= CurrentTime)  /* a gline has expired */
            {
              if(glines == kill_list_ptr)
                {
                  /* Its pointing to first one in link list*/
                  /* so, bypass this one, remember bad things can happen
                     if you try to use an already freed pointer.. */

                  glines = last_list_ptr = tmp_list_ptr =
                    kill_list_ptr->next;
                }
              else
                {
                  /* its in the middle of the list, so link around it */
                  tmp_list_ptr = last_list_ptr->next = kill_list_ptr->next;
                }

              free_conf(kill_list_ptr);
              kill_list_ptr = tmp_list_ptr;
            }
          else
            {
              if( (kill_list_ptr->user && (!username || match(kill_list_ptr->user,
                 username))) && (kill_list_ptr->host &&
                   (!host || match(kill_list_ptr->host,host))))
                return(kill_list_ptr);
              last_list_ptr = kill_list_ptr;
              kill_list_ptr = kill_list_ptr->next;
            }
        }
    }

  return((aConfItem *)NULL);
}

/* report_glines
 *
 * inputs       - aClient pointer
 * output       - NONE
 * side effects - 
 *
 * report pending glines, and placed glines.
 * 
 * - Dianora              
 */
void report_glines(aClient *sptr)
{
  aConfItem *kill_list_ptr;
  aConfItem *last_list_ptr;
  aConfItem *tmp_list_ptr;
  char *host;
  char *name;
  char reason[255];

  if(glines)
    {
      kill_list_ptr = last_list_ptr = glines;

      while(kill_list_ptr)
        {
          if(kill_list_ptr->hold <= CurrentTime)   /* gline has expired */
            {
              if(glines == kill_list_ptr)
                {
                  glines = last_list_ptr = tmp_list_ptr =
                    kill_list_ptr->next;
                }
              else
                {
                  /* its in the middle of the list, so link around it */
                  tmp_list_ptr = last_list_ptr->next = kill_list_ptr->next;
                }
              free_conf(kill_list_ptr);
              kill_list_ptr = tmp_list_ptr;
            }
          else
            {
              if(kill_list_ptr->host)
                host = kill_list_ptr->host;
              else
                host = "*";

              if(kill_list_ptr->user)
                name = kill_list_ptr->user;
              else
                name = "*";

              if(kill_list_ptr->passwd)	
				{
				  if(strlen(kill_list_ptr->passwd) > 200)	/* just some overflow prevention */
					kill_list_ptr->passwd[200]='0';
					
				  ircsprintf(reason,"\2[\2%s\2]\2 from \2%s\2 expiring on \2%s\2",
					  kill_list_ptr->passwd, kill_list_ptr->name, smalldate(kill_list_ptr->hold));
				}
              else
                strncpy_irc(reason, "No Reason", sizeof(reason));

              sendto_one(sptr,form_str(RPL_STATSKLINE), me.name,
                         sptr->name, 'G' , host, name, reason);

              last_list_ptr = kill_list_ptr;
              kill_list_ptr = kill_list_ptr->next;
            }
        }
    }
}



/* add_gline
 *
 * inputs       - pointer to aConfItem
 * output       - none
 * Side effects - links in given aConfItem into gline link list
 *
 * Identical to add_temp_kline code really.
 *
 * -Dianora
 */

static void add_gline(aConfItem *aconf)
{
  aconf->next = glines;
  glines = aconf;
}

/* send_all_glines
 *
 * inputs       - pointer to aConfItem
 * output       - none
 * Side effects - links in given aConfItem into gline link list
 *
 * Identical to add_temp_kline code really.
 *
 * -Dianora
 */
void send_all_glines(struct Client *acptr)
{
  aConfItem *aconf = glines;
  
  while(aconf)
	{
	  if(IsCapable(acptr,CAP_PTS4))
		sendto_one(acptr, ":%s GLINE %s@%s %lu %s :%s",
			me.name,
			aconf->user, aconf->host,
			aconf->hold - CurrentTime,
			aconf->name,
			aconf->passwd);
	  else
	  		sendto_one(acptr, ":%s GLINE %s@%s %lu %lu %s :%s",
			me.name,
			aconf->user, aconf->host,
			aconf->hold - CurrentTime,
			CurrentTime,
			aconf->name,
			aconf->passwd);
	  aconf = aconf->next;
	}
}

